import logging

from PyQt5 import sip
from PyQt5.QtWidgets import QWidget, QVBoxLayout

import idaapi

from libbs.ui.version import set_ui_version
set_ui_version("PyQt5")

from dailalib.llm_chat.llm_chat_ui import LLMChatClient

_l = logging.getLogger(__name__)

# disable the annoying "Running Python script" wait box that freezes IDA at times
idaapi.set_script_timeout(0)


class LLMChatWrapper(object):
    NAME = "LLM Chat"

    def __init__(self, ai_api, context=None):
        # create a dockable view
        self.twidget = idaapi.create_empty_widget(LLMChatWrapper.NAME)
        self.widget = sip.wrapinstance(int(self.twidget), QWidget)
        self.widget.name = LLMChatWrapper.NAME
        self.width_hint = 250

        self._ai_api = ai_api
        self._context = context
        self._w = None

        self._init_widgets()

    def _init_widgets(self):
        self._w = LLMChatClient(self._ai_api, context=self._context)
        layout = QVBoxLayout()
        layout.addWidget(self._w)
        layout.setContentsMargins(2,2,2,2)
        self.widget.setLayout(layout)


def add_llm_chat_to_ui(*args, ai_api=None, context=None, **kwargs):
    """
    Open the control panel view and attach it to IDA View-A or Pseudocode-A.
    """
    wrapper = LLMChatWrapper(ai_api, context=context)
    if not wrapper.twidget:
        _l.info("Unable to find a widget to attach to. You are likely running headlessly")
        return None

    flags = idaapi.PluginForm.WOPN_TAB | idaapi.PluginForm.WOPN_RESTORE | idaapi.PluginForm.WOPN_PERSIST
    idaapi.display_widget(wrapper.twidget, flags)
    wrapper.widget.visible = True
    target = "Pseudocode-A"
    dwidget = idaapi.find_widget(target)

    if not dwidget:
        target = "IDA View-A"

    idaapi.set_dock_pos(LLMChatWrapper.NAME, target, idaapi.DP_RIGHT)



import logging

from libbs.api import DecompilerInterface
from libbs.decompilers import IDA_DECOMPILER, ANGR_DECOMPILER, BINJA_DECOMPILER, GHIDRA_DECOMPILER

from ..api import AIAPI

_l = logging.getLogger(__name__)


def get_llm_chat_creator(ai_api: AIAPI) -> callable:
    # determine the current decompiler
    current_decompiler = DecompilerInterface.find_current_decompiler()
    add_llm_chat_to_ui = lambda *args, **kwargs: None
    if current_decompiler == IDA_DECOMPILER:
        from dailalib.llm_chat.ida import add_llm_chat_to_ui
    elif current_decompiler == BINJA_DECOMPILER:
        from dailalib.llm_chat.binja import add_llm_chat_to_ui
    else:
        _l.warning(f"LLM Chat not supported for decompiler %s", current_decompiler)

    def llm_chat_creator_wrapper(*args, **kwargs):
        ai_api.info(f"Opening LLM Chat with model {ai_api.model}...")
        return add_llm_chat_to_ui(ai_api=ai_api, *args, **kwargs)

    return llm_chat_creator_wrapper



import os
from pathlib import Path
import logging

from binsync.ui.qt_objects import (
    QComboBox,
    QDialog,
    QDir,
    QFileDialog,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QVBoxLayout
)
from . import AIBSUser, add_ai_user_to_project
from binsync.api.controller import BSController

_l = logging.getLogger(__name__)
AUTO_DECOMPILER = "automatic"


class AIUserConfigDialog(QDialog):
    TITLE = "AI User Configuration"

    def __init__(self, controller: BSController, parent=None):
        super().__init__(parent)
        self._controller = controller
        self.api_key = os.getenv("OPENAI_API_KEY") or ""
        self.username = f"{self._controller.client.master_user}_{AIBSUser.DEFAULT_USERNAME}"
        self.project_path = str(Path(controller.client.repo_root).absolute())
        self.binary_path = str(Path(controller.binary_path()).absolute()) if controller.binary_path() else ""
        self.base_on = ""
        self.model = "gpt-4"

        self.setWindowTitle(self.TITLE)
        self._main_layout = QVBoxLayout()
        self._grid_layout = QGridLayout()
        self.row = 0

        self._init_widgets()
        self._main_layout.addLayout(self._grid_layout)
        self.setLayout(self._main_layout)

    def _init_widgets(self):
        # model selection
        self._model_label = QLabel("AI Model")
        self._grid_layout.addWidget(self._model_label, self.row, 0)
        self._model_dropdown = QComboBox()
        # TODO: add more decompilers
        self._model_dropdown.addItems(["gpt-4", "gpt-3.5-turbo", "VARModel"])
        self._grid_layout.addWidget(self._model_dropdown, self.row, 1)
        self.row += 1

        # api key label
        self._api_key_label = QLabel("API Key")
        self._grid_layout.addWidget(self._api_key_label, self.row, 0)
        # api key input
        self._api_key_input = QLineEdit(self.api_key)
        self._grid_layout.addWidget(self._api_key_input, self.row, 1)
        self.row += 1

        # username label
        self._username_label = QLabel("Username")
        self._grid_layout.addWidget(self._username_label, self.row, 0)
        # username input
        self._username_input = QLineEdit(self.username)
        self._grid_layout.addWidget(self._username_input, self.row, 1)
        self.row += 1

        # binary label
        self._binary_path_label = QLabel("Binary Path")
        self._grid_layout.addWidget(self._binary_path_label, self.row, 0)
        # binary input
        self._binary_path_input = QLineEdit(self.binary_path)
        self._grid_layout.addWidget(self._binary_path_input, self.row, 1)
        # project button
        self._binary_path_button = QPushButton("...")
        self._binary_path_button.clicked.connect(self._on_binary_path_button_blocked)
        self._grid_layout.addWidget(self._binary_path_button, self.row, 2)
        self.row += 1

        # decompiler dropdown selection
        #self._decompiler_label = QLabel("Decompiler Backend")
        #self._grid_layout.addWidget(self._decompiler_label, self.row, 0)
        #self._decompiler_dropdown = QComboBox()
        ## TODO: add more decompilers
        #self._decompiler_dropdown.addItems([AUTO_DECOMPILER, ANGR_DECOMPILER])
        #self._grid_layout.addWidget(self._decompiler_dropdown, self.row, 1)
        #self.row += 1

        # user_base dropdown selection
        self._user_base_label = QLabel("Base On")
        self._grid_layout.addWidget(self._user_base_label, self.row, 0)
        self._user_base_dropdown = QComboBox()

        all_users = [user.name for user in self._controller.users()]
        curr_user = self._controller.client.master_user
        all_users.remove(curr_user)
        all_users = [""] + [curr_user] + all_users
        self._user_base_dropdown.addItems(all_users)
        self._grid_layout.addWidget(self._user_base_dropdown, self.row, 1)
        self.row += 1

        # range selection
        self._range_label = QLabel("View Range")
        self._grid_layout.addWidget(self._range_label, self.row, 0)
        self._range_input = QLineEdit("")
        self._grid_layout.addWidget(self._range_input, self.row, 1)
        self.row += 1

        # ok/cancel buttons
        self._ok_button = QPushButton("OK")
        self._ok_button.clicked.connect(self._on_ok_button_clicked)
        self._cancel_button = QPushButton("Cancel")
        self._cancel_button.clicked.connect(self._on_cancel_button_clicked)
        self._button_layout = QHBoxLayout()
        self._button_layout.addWidget(self._ok_button)
        self._button_layout.addWidget(self._cancel_button)

        self._main_layout.addLayout(self._grid_layout)
        self._main_layout.addLayout(self._button_layout)

    def _on_binary_path_button_blocked(self):
        # get the path to the binary
        binary_path = QFileDialog.getOpenFileName(self, "Select Binary", QDir.homePath())
        if binary_path[0]:
            self._binary_path_input.setText(binary_path[0])

    def _on_ok_button_clicked(self):
        self.api_key = self._api_key_input.text()
        self.binary_path = self._binary_path_input.text()
        self.username = self._username_input.text()
        #self.decompiler_backend = self._decompiler_dropdown.currentText()
        self.decompiler_backend = AUTO_DECOMPILER
        if self.decompiler_backend == AUTO_DECOMPILER:
            self.decompiler_backend = None

        self.base_on = self._user_base_dropdown.currentText()
        self.model = self._model_dropdown.currentText()
        self.range_str = self._range_input.text()

        if not (self.api_key and self.binary_path and self.username):
            if self.model is not None and "gpt" in self.model:
                _l.critical("You did not provide a path, username, and API key for the AI user.")
                return

        _l.info(f"Starting AI user now! Commits from user {self.username} should appear soon...")
        self.hide()
        try:
            self.threaded_add_ai_user_to_project()
        except Exception as e:
            _l.info(f"Ran into issue: {e}")

        self.close()

    def _on_cancel_button_clicked(self):
        self.close()
    
    def threaded_add_ai_user_to_project(self):
        # angr hack to make sure the workspace is visible!
        arch = None
        if hasattr(self._controller, "workspace"):
            globals()['workspace'] = self._controller.workspace
            arch = self._controller.main_instance.project.arch.name

        add_ai_user_to_project(
            self.api_key, self.binary_path, self.project_path, username=self.username,
            base_on=self.base_on, headless=True if self.decompiler_backend else False, copy_proj=True, model=self.model,
            decompiler_backend=self.decompiler_backend, range_str=self.range_str, arch=arch
        )
        self._controller.used_ai_user = True


from binaryninjaui import (
    UIContext,
    DockHandler,
    DockContextHandler,
    UIAction,
    UIActionHandler,
    Menu,
)

import traceback
import sys
import logging

from PySide6.QtWidgets import (
    QDockWidget,
    QWidget,
    QApplication,
    QMenu,
    QMainWindow,
    QMenuBar, QVBoxLayout,
)
from PySide6.QtCore import Qt
from binaryninjaui import DockContextHandler
import binaryninja

from .llm_chat_ui import LLMChatClient
from ..api import AIAPI

_l = logging.getLogger(__name__)

# this code is based on:
# https://github.com/binsync/binsync/blob/fa754795b7d55e4b5de4e12ea7d4b5f9706c23af/plugins/binja_binsync/binja_binsync.py

def find_main_window():
    main_window = None
    for x in QApplication.allWidgets():
        if not isinstance(x, QDockWidget):
            continue
        main_window = x.parent()
        if isinstance(main_window, (QMainWindow, QWidget)):
            break
        else:
            main_window = None

    if main_window is None:
        # oops cannot find the main window
        raise Exception("Main window is not found.")
    return main_window


dockwidgets = [ ]


# shamelessly copied from https://github.com/Vector35/debugger
def create_widget(widget_class, name, parent, data, *args):
    # It is imperative this function return *some* value because Shiboken will try to deref what we return
    # If we return nothing (or throw) there will be a null pointer deref (and we won't even get to see why)
    # So in the event of an error or a nothing, return an empty widget that at least stops the crash
    try:
        # binsync specific code
        if not isinstance(data, binaryninja.BinaryView):
            raise Exception('expected an binary view')
        new_bv = args[0]
        # uses only a bv_controller
        widget = widget_class(new_bv, parent=parent, name=name, data=data)
        if not widget:
            raise Exception('expected widget, got None')

        global dockwidgets

        found = False
        for (bv, widgets) in dockwidgets:
            if bv == data:
                widgets[name] = widget
                found = True

        if not found:
            dockwidgets.append((data, {
                name: widget
            }))

        widget.destroyed.connect(lambda destroyed: destroy_widget(destroyed, widget, data, name))

        return widget
    except Exception:
        traceback.print_exc(file=sys.stderr)
        return QWidget(parent)


def destroy_widget(destroyed, old, data, name):
    # Gotta be careful to delete the correct widget here
    for (bv, widgets) in dockwidgets:
        if bv == data:
            for (name, widget) in widgets.items():
                if widget == old:
                    # If there are no other references to it, this will be the only one and the call
                    # will delete it and invoke __del__.
                    widgets.pop(name)
                    return


class BinjaWidgetBase:
    def __init__(self):
        self._main_window = None
        self._menu_bar = None
        self._tool_menu = None

    @property
    def main_window(self):
        if self._main_window is None:
            self._main_window = find_main_window()
        return self._main_window

    @property
    def menu_bar(self):
        if self._menu_bar is None:
            self._menu_bar = next(
                iter(x for x in self._main_window.children() if isinstance(x, QMenuBar))
            )
        return self._menu_bar

    @property
    def tool_menu(self):
        if self._tool_menu is None:
            self._tool_menu = next(
                iter(
                    x
                    for x in self._menu_bar.children()
                    if isinstance(x, QMenu) and x.title() == u"Tools"
                )
            )
        return self._tool_menu

    def add_tool_menu_action(self, name, func):
        self.tool_menu.addAction(name, func)


class BinjaDockWidget(QWidget, DockContextHandler):
    def __init__(self, name, parent=None):
        QWidget.__init__(self, parent)
        DockContextHandler.__init__(self, self, name)

        self.base = BinjaWidgetBase()

        self.show()

    def toggle(self):
        if self.isVisible():
            self.hide()
        else:
            self.show()


class BinjaWidget(QWidget):
    def __init__(self, tabname):
        super(BinjaWidget, self).__init__()


class LLMChatClientDockWidget(BinjaDockWidget):
    def __init__(self, ai_api, parent=None, name=None, data=None):
        super().__init__(name, parent=parent)
        self.data = data
        self._widget = None
        self.ai_api = ai_api

        self._widget = LLMChatClient(self.ai_api)
        layout = QVBoxLayout()
        layout.addWidget(self._widget)
        self.setLayout(layout)


def add_llm_chat_to_ui(*args, ai_api: AIAPI = None, **kwargs):
    # control panel (per BV)
    dock_handler = DockHandler.getActiveDockHandler()
    dock_handler.addDockWidget(
        "LLM Chat",
        lambda n, p, d: create_widget(LLMChatClientDockWidget, n, p, d, ai_api._dec_interface.bv),
        Qt.RightDockWidgetArea,
        Qt.Vertical,
        True
    )


from typing import Optional
import os

import tiktoken

from ..ai_api import AIAPI

active_model = None
active_prompt_style = None


class LiteLLMAIAPI(AIAPI):
    prompts_by_name = []
    DEFAULT_MODEL = "gpt-4o"
    OPENAI_MODELS = {"gpt-4", "gpt-4o", "gpt-4-turbo", "gpt-3.5-turbo", "o1-mini", "o1-preview"}
    MODEL_TO_TOKENS = {
        # TODO: update the token values for o1
        "o1-mini": 8_000,
        "o1-preview": 8_000,
        "gpt-4o": 8_000,
        "gpt-4o-mini": 16_000,
        "gpt-4-turbo": 128_000,
        "claude-3-5-sonnet-20240620": 200_000,
        "gemini/gemini-pro": 12_288,
        "vertex_ai_beta/gemini-pro": 12_288,
    }

    # replacement strings for API calls
    def __init__(
        self,
        api_key: Optional[str] = None,
        model: str = DEFAULT_MODEL,
        prompt_style: str = "few-shot",
        prompts: Optional[list] = None,
        fit_to_tokens: bool = False,
        chat_use_ctx: bool = True,
        chat_event_callbacks: Optional[dict] = None,
        **kwargs
    ):
        super().__init__(**kwargs)
        self._api_key = None
        # default to openai api key if not provided
        self.api_key = api_key or os.getenv("OPENAI_API_KEY")
        self.model = model
        self.prompt_style = prompt_style
        self.fit_to_tokens = fit_to_tokens
        self.chat_use_ctx = chat_use_ctx
        self.chat_event_callbacks = chat_event_callbacks or {"send": None, "receive": None}

        # delay prompt import
        from .prompts import PROMPTS
        prompts = prompts + PROMPTS if prompts else PROMPTS
        self.prompts_by_name = {p.name: p for p in prompts}

        # update the globals (for threading hacks)
        global active_model, active_prompt_style
        active_model = self.model
        active_prompt_style = self.prompt_style

    def __dir__(self):
        return list(super().__dir__()) + list(self.prompts_by_name.keys())

    def __getattribute__(self, item):
        # this is how we can access the prompt functions
        if item in object.__getattribute__(self, "prompts_by_name"):
            prompt_obj: "Prompt" = self.prompts_by_name[item]
            prompt_obj.ai_api = self
            return prompt_obj.query_model
        else:
            return object.__getattribute__(self, item)

    def query_model(
        self,
        prompt: str,
        model: Optional[str] = None,
        max_tokens=None,
    ):
        # delay import because litellm attempts to query the server on import to collect cost information.
        from litellm import completion

        if not self.api_key:
            raise ValueError(f"Model API key is not set. Please set it before querying the model {self.model}")

        prompt_model = model or self.model
        response = completion(
            model=prompt_model,
            messages=[
                {"role": "user", "content": prompt}
            ],
            max_tokens=max_tokens,
            timeout=60,
        )
        # get the answer
        try:
            answer = response.choices[0].message.content
        except (KeyError, IndexError) as e:
            answer = None

        # get the estimated cost
        try:
            prompt_tokens = response.usage.prompt_tokens
            completion_tokens = response.usage.completion_tokens
        except (KeyError, IndexError) as e:
            prompt_tokens, completion_tokens = None, None
        cost = self.llm_cost(prompt_model, prompt_tokens, completion_tokens) \
            if prompt_tokens is not None and completion_tokens is not None else None

        return answer, cost

    @staticmethod
    def estimate_token_amount(content: str, model=DEFAULT_MODEL):
        # TODO: we only support one token counting for now, gpt-4
        enc = tiktoken.encoding_for_model("gpt-4")
        tokens = enc.encode(content)
        return len(tokens)

    @staticmethod
    def content_fits_tokens(content: str, model=DEFAULT_MODEL):
        max_token_count = LiteLLMAIAPI.MODEL_TO_TOKENS[model]
        token_count = LiteLLMAIAPI.estimate_token_amount(content, model=model)
        return token_count <= max_token_count - 1000

    @staticmethod
    def fit_decompilation_to_token_max(decompilation: str, delta_step=10, model=DEFAULT_MODEL):
        if LiteLLMAIAPI.content_fits_tokens(decompilation, model=model):
            return decompilation

        dec_lines = decompilation.split("\n")
        last_idx = len(dec_lines) - 1
        # should be: [func_prototype] + [nop] + [mid] + [nop] + [end_of_code]
        dec_lines = dec_lines[0:2] + ["// ..."] + dec_lines[delta_step:last_idx-delta_step] + ["// ..."] + dec_lines[-2:-1]
        decompilation = "\n".join(dec_lines)

        return LiteLLMAIAPI.fit_decompilation_to_token_max(decompilation, delta_step=delta_step, model=model)

    @staticmethod
    def llm_cost(model_name: str, prompt_tokens: int, completion_tokens: int) -> float | None:
        # these are the $ per million tokens
        COST = {
            "gpt-4o": {"prompt_price": 2.5, "completion_price": 10},
            "gpt-4o-mini": {"prompt_price": 0.150, "completion_price": 0.600},
            "gpt-4-turbo": {"prompt_price": 10, "completion_price": 30},
            "claude-3.5-sonnet-20240620": {"prompt_price": 3, "completion_price": 15},
            "gemini/gemini-pro": {"prompt_price": 0.150, "completion_price": 0.600},
            "vertex_ai_beta/gemini-pro": {"prompt_price": 0.150, "completion_price": 0.600},
        }
        if model_name not in COST:
            return None

        llm_price = COST[model_name]
        prompt_price = (prompt_tokens / 1000000) * llm_price["prompt_price"]
        completion_price = (completion_tokens / 1000000) * llm_price["completion_price"]

        return round(prompt_price + completion_price, 5)

    #
    # LMM Settings
    #

    @property
    def api_key(self):
        if not self._api_key or self.model is None:
            return None
        elif self.model in self.OPENAI_MODELS:
            return os.getenv("OPENAI_API_KEY", None)
        elif "claude" in self.model:
            return os.getenv("ANTHROPIC_API_KEY", None)
        elif "gemini/gemini" in self.model:
            return os.getenv("GEMINI_API_KEY", None)
        elif "vertex" in self.model:
            return self._api_key
        else:
            return None

    @api_key.setter
    def api_key(self, value):
        self._api_key = value
        if self._api_key and self.model is not None:
            if self.model in self.OPENAI_MODELS:
                os.environ["OPENAI_API_KEY"] = self._api_key
            elif "claude" in self.model:
                os.environ["ANTHROPIC_API_KEY"] = self._api_key
            elif "gemini/gemini" in self.model:
                os.environ["GEMINI_API_KEY"] = self._api_key

    def ask_api_key(self, *args, **kwargs):
        api_key_or_path = self._dec_interface.gui_ask_for_string("Enter you AI API Key or Creds Path:", title="DAILA")
        if "/" in api_key_or_path or "\\" in api_key_or_path:
            # treat as path
            with open(api_key_or_path, "r") as f:
                api_key = f.read().strip()
        else:
            api_key = api_key_or_path
        self.api_key = api_key

    def _set_prompt_style(self, prompt_style):
        self.prompt_style = prompt_style
        global active_prompt_style
        active_prompt_style = prompt_style

    def _set_model(self, model):
        self.model = model
        global active_model
        active_model = model

    def get_model(self):
        # TODO: this hack needs to be refactored later
        global active_model
        return str(active_model)

    def ask_prompt_style(self, *args, **kwargs):
        if self._dec_interface is not None:
            from .prompts import ALL_STYLES

            prompt_style = self.prompt_style
            style_choices = ALL_STYLES.copy()
            if self.prompt_style:
                style_choices.remove(self.prompt_style)
                style_choices = [self.prompt_style] + style_choices

            p_style = self._dec_interface.gui_ask_for_choice(
                "What prompting style would you like to use?",
                style_choices,
                title="DAILA"
            )
            if p_style != prompt_style and p_style:
                if p_style not in ALL_STYLES:
                    self._dec_interface.error(f"Prompt style {p_style} is not supported.")
                    return

                self._set_prompt_style(p_style)
                self._dec_interface.info(f"Prompt style set to {p_style}")

    def ask_model(self, *args, **kwargs):
        if self._dec_interface is not None:
            model_choices = list(LiteLLMAIAPI.MODEL_TO_TOKENS.keys())
            if self.model:
                model_choices.remove(self.model)
                model_choices = [self.model] + model_choices

            model = self._dec_interface.gui_ask_for_choice(
                "What LLM model would you like to use?",
                model_choices,
                title="DAILA"
            )
            self._set_model(model)
            self._dec_interface.info(f"Model set to {model}")




import typing

from PyQt5.QtWidgets import (
    QApplication, QWidget, QVBoxLayout, QHBoxLayout, QTextEdit,
    QPushButton, QLabel, QScrollArea, QFrame
)
from PyQt5.QtCore import Qt, QThread, pyqtSignal, QCoreApplication
from PyQt5.QtGui import QFont

from libbs.artifacts.context import Context

if typing.TYPE_CHECKING:
    from ..api.litellm.litellm_api import LiteLLMAIAPI

CONTEXT_PROMPT = """
You are reverse engineering assistant that helps to understand binaries in a decompiler. Given decompilation 
and questions you answer them to the best of your ability. Here is the function you are currently working on:
```
DEC_TEXT
```

Acknowledging the context, by responding with:
"I see you are working on function <function_name>. How can I help you today?"
"""


class LLMChatClient(QWidget):
    def __init__(self, ai_api: "LiteLLMAIAPI", parent=None, context: Context = None):
        super(LLMChatClient, self).__init__(parent)
        self.model = ai_api.get_model()
        self.ai_api = ai_api
        self.context = context
        self.setWindowTitle('LLM Chat')
        self.setGeometry(100, 100, 600, 800)

        # Main layout
        self.layout = QVBoxLayout(self)
        self.setLayout(self.layout)

        # Scroll area for chat messages
        self.chat_area = QScrollArea()
        self.chat_area.setWidgetResizable(True)
        self.chat_content = QWidget()
        self.chat_layout = QVBoxLayout(self.chat_content)
        self.chat_layout.addStretch(1)
        self.chat_area.setWidget(self.chat_content)

        # Input area
        self.input_text = QTextEdit()
        self.input_text.setFixedHeight(80)

        # Send button
        self.send_button = QPushButton('Send')
        self.send_button.setFixedHeight(40)
        self.send_button.clicked.connect(lambda: self.send_message())

        # Arrange input and send button horizontally
        self.input_layout = QHBoxLayout()
        self.input_layout.addWidget(self.input_text)
        self.input_layout.addWidget(self.send_button)

        # Add widgets to the main layout
        self.layout.addWidget(self.chat_area)
        self.layout.addLayout(self.input_layout)

        # Chat history
        self.chat_history = []

        # model check
        if not self.model:
            self.ai_api.warning("No model set. Close the chat window and please set a model before using the chat")
            return

        # preset the very first interaction
        # create a context for this first message
        if ai_api.chat_use_ctx:
            ai_api.info("Collecting context for the current function...")
            if context is None:
                context = ai_api._dec_interface.gui_active_context()
            dec = ai_api._dec_interface.decompile(context.func_addr)
            dec_text = dec.text if dec is not None else None
            if dec_text:
                # put a number in front of each line
                dec_lines = dec_text.split("\n")
                dec_text = "\n".join([f"{i + 1} {line}" for i, line in enumerate(dec_lines)])
                prompt = CONTEXT_PROMPT.replace("DEC_TEXT", dec_text)
                # set the text to the prompt
                self.input_text.setText(prompt)
                self.send_message(add_text=False, role="system")
        else:
            self.input_text.setText("Say how can I help you.")
            self.send_message(add_text=False, role="system")

    def add_message(self, text, is_user):
        # Message bubble
        message_label = QLabel(text)
        message_label.setWordWrap(True)
        message_label.setFont(QFont('Arial', 12))
        message_label.setTextInteractionFlags(Qt.TextSelectableByMouse)

        # Bubble styling
        bubble = QFrame()
        bubble_layout = QHBoxLayout()
        bubble.setLayout(bubble_layout)

        if is_user:
            # User message on the right
            message_label.setStyleSheet("""
                background-color: #DCF8C6;
                color: black;
                padding: 10px;
                border-radius: 10px;
            """)
            bubble_layout.addStretch()
            bubble_layout.addWidget(message_label)
        else:
            # Assistant message on the left
            message_label.setStyleSheet("""
                background-color: #FFFFFF;
                color: black;
                padding: 10px;
                border-radius: 10px;
            """)
            bubble_layout.addWidget(message_label)
            bubble_layout.addStretch()

        self.chat_layout.insertWidget(self.chat_layout.count() - 1, bubble)
        QCoreApplication.processEvents()
        self.chat_area.verticalScrollBar().setValue(self.chat_area.verticalScrollBar().maximum())

    def send_message(self, add_text=True, role="user"):
        user_text = self.input_text.toPlainText().strip()
        if not user_text:
            return

        # do aiapi calback
        if self.ai_api:
            send_callback = self.ai_api.chat_event_callbacks.get("send", None)
            if send_callback:
                send_callback(user_text)

        # Display user message
        if add_text:
            self.add_message(user_text, is_user=True)
        self.input_text.clear()

        # Append to chat history
        self.chat_history.append({"role": role, "content": user_text})

        # Disable input while waiting for response
        self.input_text.setDisabled(True)
        self.send_button.setDisabled(True)

        # Start a thread to get the response
        self.thread = LLMThread(self.chat_history, self.model)
        self.thread.response_received.connect(lambda msg: self.receive_message(msg))
        self.thread.start()

    def receive_message(self, assistant_message):
        # Display assistant message
        self.add_message(assistant_message, is_user=False)

        # do aiapi calback
        if self.ai_api:
            recv_callback = self.ai_api.chat_event_callbacks.get("receive", None)
            if recv_callback:
                recv_callback(assistant_message)

        # Append to chat history
        self.chat_history.append({"role": "user", "content": assistant_message})

        # Re-enable input
        self.input_text.setDisabled(False)
        self.send_button.setDisabled(False)

    def closeEvent(self, event):
        # Ensure that the thread is properly terminated when the window is closed
        if hasattr(self, 'thread') and self.thread.isRunning():
            self.thread.terminate()
        event.accept()


class LLMThread(QThread):
    response_received = pyqtSignal(str)

    def __init__(self, chat_history, model_name):
        super().__init__()
        self.chat_history = chat_history.copy()
        self.model_name = model_name

    def run(self):
        from litellm import completion

        response = completion(
            model=self.model_name,
            messages=self.chat_history,
            timeout=60,
        )

        try:
            answer = response.choices[0].message.content
        except (KeyError, IndexError) as e:
            answer = f"Error: {e}. Please close the window and try again."

        self.response_received.emit(answer)



from pathlib import Path
import argparse
import subprocess
import sys

from dailalib.binsync_plugin.ai_bs_user import AIBSUser
from dailalib.binsync_plugin.openai_bs_user import OpenAIBSUser
from dailalib.binsync_plugin.varmodel_bs_user import VARModelBSUser


def add_ai_user_to_project(
    openai_api_key: str, binary_path: Path, bs_proj_path: Path, username: str = AIBSUser.DEFAULT_USERNAME,
    base_on=None, headless=False, copy_proj=False, decompiler_backend=None, model=None, controller=None, progress_callback=None,
    range_str="", arch=None
):
    if headless:
        _headlessly_add_ai_user(openai_api_key, binary_path, bs_proj_path, username=username, decompiler_backend=decompiler_backend, base_on=base_on, model=model, arch=arch)
    else:
        if model is None or model.startswith("gpt"):
            ai_user = OpenAIBSUser(
                openai_api_key=openai_api_key, binary_path=binary_path, bs_proj_path=bs_proj_path, model=model,
                username=username, copy_project=copy_proj, decompiler_backend=decompiler_backend, base_on=base_on,
                controller=controller, progress_callback=progress_callback, range_str=range_str, arch=arch
            )
        elif model == "VARModel":
            ai_user = VARModelBSUser(
                openai_api_key=openai_api_key, binary_path=binary_path, bs_proj_path=bs_proj_path, model=model,
                username=username, copy_project=copy_proj, decompiler_backend=decompiler_backend, base_on=base_on,
                controller=controller, progress_callback=progress_callback, range_str=range_str, arch=arch
            )
        else:
            raise ValueError(f"Model: {model} is not supported. Please use a supported model.")

        ai_user.add_ai_user_to_project()


def _headlessly_add_ai_user(
    openai_api_key: str, binary_path: Path, bs_proj_path: Path, username: str = AIBSUser.DEFAULT_USERNAME,
    decompiler_backend=None, base_on=None, model=None, arch=None
):
    script_path = Path(__file__).absolute()
    python_path = sys.executable
    optional_args = []
    if decompiler_backend:
        optional_args += ["--dec", decompiler_backend]
    if base_on:
        optional_args += ["--base-on", base_on]
    if model:
        optional_args += ["--model", model]
    if arch:
        optional_args += ["--arch", arch]

    subpproc = subprocess.Popen([
        python_path,
        str(script_path),
        openai_api_key,
        str(binary_path),
        "--username",
        username,
        "--proj-path",
        str(bs_proj_path),
    ] + optional_args)
    return subpproc


def _headless_main():
    parser = argparse.ArgumentParser()
    parser.add_argument("openai_api_key", type=str)
    parser.add_argument("binary_path", type=Path)
    parser.add_argument("--proj-path", type=Path)
    parser.add_argument("--username", type=str)
    parser.add_argument("--dec", type=str)
    parser.add_argument("--base-on", type=str)
    parser.add_argument("--model", type=str)
    parser.add_argument("--arch", type=str)

    args = parser.parse_args()
    if args.username is None:
        args.username = AIBSUser.DEFAULT_USERNAME

    add_ai_user_to_project(
        args.openai_api_key, args.binary_path, args.proj_path, username=args.username, headless=False,
        copy_proj=True, decompiler_backend=args.dec if args.dec else None, base_on=args.base_on,
        model=args.model if args.model else None, arch=args.arch if args.arch else None
    )


if __name__ == "__main__":
    _headless_main()



RENAME_FUNCTION = """
# Task
You are a decompiled C expert that renames functions. When given a function, you rename it according to the
meaning of the function or its use. You specify which function you are renaming by its name.

You eventually respond with a valid json. As an example:
## Answer 
{
    "sub_404000": "fibonacci",
}

{% if (few_shot or cot) %}
# Example
Here is an example. Given the following code:
```
int sub_404000(int a0, char** a1)
{
    int is_even; // rax

    is_even = sub_404100(a0[1]) % 2 == 0
    return is_even;
}
```

You respond with:
## Answer
{
    "sub_404000": "is_even",
    "sub_404100": "get_value",
}
{% endif %}

# Example
Given the following code:
```
{{ decompilation }}
```

You respond with:
"""

RENAME_VARIABLES = """
# Task
You are decompiled C expert that renames variables in code. When given code, you rename variables according to the
meaning of the function or its use.

You eventually respond with a valid json. As an example:
## Answer 
{
  "v1": "i",
  "v2": "ptr"
}

{% if few_shot %}
# Example
Here is an example. Given the following code:
```
int sub_404000(int a0, char** a1)
{
    int v1; // rax

    v1 = sub_404100(a0[1]) % 2 == 0
    return v1;
}
```

You responded with:
## Answer
{
    "a0": "argc",
    "a1": "argv",
    "v1": "is_even"
}
{% endif %}

# Example
Given the following code:
```
{{ decompilation }}
```

You respond with:
"""

SUMMARIZE_FUNCTION = """
# Task
You are decompiled C expert that summarizes code. When given code, you summarize at a high level what the function does
and you identify if known algorithms are used in the function.

You eventually respond with a valid json. As an example:
## Answer 
{
    "summary": "This function computes the fibonacci sequence. It takes an integer as an argument and returns the fibonacci number at that index.",
    "algorithms": ["fibonacci"]
}

{% if few_shot %}
# Example
Here is an example. Given the following code:
```
int sub_404000(int a0, char** a1)
{
    int v1; // rax
    v1 = sub_404100(a0[1]) % 2 == 0
    return v1;
}
```

You responded with:
## Answer
{
    "summary": "This function takes two arguments and implements the is_even check on second argument",
    "algorithms": ["is_even"]
}
{% endif %}

# Example
Given the following code:
```
{{ decompilation }}
```

You respond with:
"""

IDENTIFY_SOURCE = """
# Task
You are a decompiled C expert that identifies the original source given decompilation. Upon discovering the source,
you give a link to the code.

You eventually respond with a valid json. As an example:
## Answer 
{
  "link": "https://github.com/torvalds/linux"
  "version": "5.10"
}

{% if few_shot %}
# Example
Here is an example. Given the following code:
```
void __fastcall __noreturn usage(int status)
{
    v2 = program_name;
    if ( status )
    {
        v3 = dcgettext(0LL, "Try '%s --help' for more information.\n", 5);
        _fprintf_chk(stderr, 1LL, v3, v2);
    }
    // ...
}
```

You would respond with:
## Answer
{
    "link": "https://www.gnu.org/software/coreutils/"
    "version": ""
}
{% endif %}

# Example
Given the following code:
```
{{ decompilation }}
```

You respond with:
"""

FIND_VULN = """
# Task
You are a decompiled C expert that identifies vulnerabilities or bugs in code. When given code, you identify 
vulnerabilities and specify the type of vulnerability. Only identify the MOST important vulnerabilities in the code.
Ignore bugs like resource leaks.

You eventually respond with a valid json. As an example:
## Answer 
{
    "vulnerabilities": ["command-injection (10-11)"],
    "description": "The function is vulnerable to a command injection on line 10 in the call to system."
}

{% if few_shot %}
# Example 
Here is an example. Given the following code:
```
1 int __fastcall __noreturn main(int argc, const char **argv, const char **envp)
2 {
3   Human *v3; // rbx
4   __int64 v4; // rdx
5   Human *v5; // rbx
6   int v6; // eax
7   __int64 v7; // rax
8   Human *v8; // rbx
9   Human *v9; // rbx
10   char v10[16]; // [rsp+10h] [rbp-50h] BYREF
11   char v11[8]; // [rsp+20h] [rbp-40h] BYREF
12   Human *v12; // [rsp+28h] [rbp-38h]
13   Human *v13; // [rsp+30h] [rbp-30h]
14   size_t nbytes; // [rsp+38h] [rbp-28h]
15   void *buf; // [rsp+40h] [rbp-20h]
16   int v16; // [rsp+48h] [rbp-18h] BYREF
17   char v17; // [rsp+4Eh] [rbp-12h] BYREF
18   char v18[17]; // [rsp+4Fh] [rbp-11h] BYREF
19
20   std::allocator<char>::allocator(&v17, argv, envp);
21   std::string::string(v10, "Jack", &v17);
22   v3 = (Human *)operator new(0x18uLL);
23   Man::Man(v3, v10, 25LL);
24   v12 = v3;
25   std::string::~string((std::string *)v10);
26   std::allocator<char>::~allocator(&v17);
27   std::allocator<char>::allocator(v18, v10, v4);
28   std::string::string(v11, "Jill", v18);
29   v5 = (Human *)operator new(0x18uLL);
30   Woman::Woman(v5, v11, 21LL);
31   v13 = v5;
32   std::string::~string((std::string *)v11);
33   std::allocator<char>::~allocator(v18);
34   while ( 1 )
35   {
36     while ( 1 )
37     {
38       while ( 1 )
39       {
40         std::operator<<<std::char_traits<char>>(&std::cout, "1. use
41 2. after
42 3. free
43 ");
44         std::istream::operator>>(&std::cin, &v16);
45         if ( v16 != 2 )
46           break;
47         nbytes = atoi(argv[1]);
48         buf = (void *)operator new[](nbytes);
49         v6 = open(argv[2], 0);
50         read(v6, buf, nbytes);
51         v7 = std::operator<<<std::char_traits<char>>(&std::cout, "your data is allocated");
52         std::ostream::operator<<(v7, &std::endl<char,std::char_traits<char>>);
53       }
54       if ( v16 == 3 )
55         break;
56       if ( v16 == 1 )
57       {
58         (*(void (__fastcall **)(Human *))(*(_QWORD *)v12 + 8LL))(v12);
59         (*(void (__fastcall **)(Human *))(*(_QWORD *)v13 + 8LL))(v13);
60       }
61     }
62     v8 = v12;
63     if ( v12 )
64     {
65       Human::~Human(v12);
66       operator delete(v8);
67     }
68     v9 = v13;
69     if ( v13 )
70     {
71       Human::~Human(v13);
72       operator delete(v9);
73     }
74   }
75 }
```

You would respond with:
## Answer
{
    "vulnerabilities": ["use-after-free (62-73)", "buffer-overflow (47-50)"]
    "description": "The code contains a classic use-after-free vulnerability. In lines 62-73, the pointers v12 and v13 (which point to objects of type Human) are deleted (freed) using operator delete. If the program's loop (lines 34-74) executes again and the pointers v12 or v13 are accessed without reallocation, it results in undefined behavior due to use-after-free. In lines 47-50, the code reads a size value from argv[1] and uses it directly with operator new[] to allocate a buffer (buf). There are no checks to ensure that nbytes is a reasonable size, potentially leading to a large allocation or integer overflow."
}
{% endif %}

# Example
Given the following code:
```
{{ decompilation }}
```

You respond with:
"""

MAN_PAGE = """
# Task
You are a decompiled C expert that generates summarized man pages for functions. You are given the function the target
function is called in for context. You generate a summarized man page for the target function. 
You only do this task for functions that are from libraries or the stdlib. Do not do this on user-defined functions.

You eventually respond with a valid json. As an example:
## Answer 
{
    "function": "printf
    "args": ["format (char *)", "arg1 (void)", "arg2 (void)"],
    "return": "int",
    "description": "The printf() function writes output to stdout using the text formatting instructions contained in the format string. The format string can contain plain text and format specifiers that begin with the % character. Each format specifier is replaced by the value of the corresponding argument in the argument list. The printf() function returns the number of characters written to stdout."
}


# Example
Given the following code as context:
```
void __fastcall gz_error(__int64 a1, int a2, const char *a3)
{
  void *v5; // rcx
  __int64 v7; // rbx
  __int64 v8; // rax
  __int64 v9; // rcx
  char *v10; // rax
  char *v11; // rcx
  const char *v12; // r9
  __int64 v13; // rax

  v5 = *(void **)(a1 + 120);
  if ( v5 )
  {
    if ( *(_DWORD *)(a1 + 116) != -4 )
      free(v5);
    *(_QWORD *)(a1 + 120) = 0LL;
  }
  if ( a2 && a2 != -5 )
    *(_DWORD *)a1 = 0;
  *(_DWORD *)(a1 + 116) = a2;
  if ( a3 && a2 != -4 )
  {
    v7 = -1LL;
    v8 = -1LL;
    do
      ++v8;
    while ( *(_BYTE *)(*(_QWORD *)(a1 + 32) + v8) );
    v9 = -1LL;
    do
      ++v9;
    while ( a3[v9] );
    v10 = (char *)malloc(v8 + 3 + v9);
    *(_QWORD *)(a1 + 120) = v10;
    v11 = v10;
    if ( v10 )
    {
      v12 = *(const char **)(a1 + 32);
      v13 = -1LL;
      while ( v12[++v13] != 0 )
        ;
      do
        ++v7;
      while ( a3[v7] );
      snprintf(v11, v7 + v13 + 3, "%s%s%s", v12, ": ", a3);
    }
    else
    {
      *(_DWORD *)(a1 + 116) = -4;
    }
  }
}
```

You focus on the line in the above text:
```
      snprintf(v11, v7 + v13 + 3, "%s%s%s", v12, ": ", a3);
```

Focusing on the outermost function call in this line, you respond with:
## Answer
{
    "function": "snprintf",
    "args": ["str (char *)", "size (size_t)", "format (const char *)", "..."],
    "return": "int",
    "description": "The snprintf() function formats and stores a series of characters and values in the array buffer. It is similar to printf(), but with two major differences: it outputs to a buffer rather than stdout, and it takes an additional size parameter specifying the limit of characters to write. The size parameter prevents buffer overflows. It returns the number of characters that would have been written if the buffer was sufficiently large, not counting the terminating null character."
}

# Example
Given the following code as context:
```
{{ decompilation }} 
```

You focus on the line in the above text:
```
{{ line_text }}
```

Focusing on the outermost function call in this line, you respond with:
"""



class PromptType:
    ZERO_SHOT = "zero-shot"
    FEW_SHOT = "few-shot"
    COT = "chain-of-thought"


ALL_STYLES = [PromptType.ZERO_SHOT, PromptType.FEW_SHOT, PromptType.COT]
DEFAULT_STYLE = PromptType.FEW_SHOT



import argparse

from .installer import DAILAInstaller
import dailalib


def main():
    parser = argparse.ArgumentParser(
        description="""
            The DAILA CLI is used to install, run, and host the DAILA plugin.
            """,
        epilog="""
            Examples:
            daila install
            """
    )
    parser.add_argument(
        "-i", "--install", action="store_true", help="Install DAILA into your decompilers"
    )
    parser.add_argument(
        "--single-decompiler-install", nargs=2, metavar=('decompiler', 'path'), help="Install DAILA into a single decompiler. Decompiler must be one of: ida, ghidra, binja, angr."
    )
    parser.add_argument(
        "-s", "--server", help="Run a a headless server for DAILA", choices=["ghidra"]
    )
    parser.add_argument(
        "-v", "--version", action="version", version=f"{dailalib.__version__}"
    )
    args = parser.parse_args()

    if args.single_decompiler_install:
        decompiler, path = args.single_decompiler_install
        DAILAInstaller().install(interactive=False, paths_by_target={decompiler: path})
    elif args.install:
        DAILAInstaller().install()
    elif args.server:
        if args.server != "ghidra":
            raise NotImplementedError("Only Ghidra is supported for now")

        from dailalib import create_plugin
        create_plugin(force_decompiler="ghidra")


if __name__ == "__main__":
    main()
